import multiprocessing
import os
import threading
import shutil
import time
from typing import Any, List, Union

import mido

from .utils import ExternalProcess, Popen, JackClient, get_smf_duration, progressbar


class MIDIPlayer(JackClient):

    MIDI_PORT = 'Carla'

    def __init__(self):
        """
        Creates a player which is able to connect to a Carla instance

        Optionally, `freewheel` can be used to start freeewheel mode before of
        playing.

        For now, only one Carla instance should be active.
        """
        super().__init__("MIDIPlayer")

    def activate(self):
        """
        Activate the MIDI player client and set the connections.

        If the Carla instance is not found, this method rase a
        `RuntimeWarning`. To avoid it, use ``Carla.exists`` method. Note that
        ``Carla.start`` already does that!
        """
        carla_ports = self.client.get_ports(self.MIDI_PORT,
                                            is_midi=True,
                                            is_input=True)
        if len(carla_ports) != 1:
            raise RuntimeWarning(
                "Cannot find the Carla instance, Retry later!")
        self.client.activate()
        self.client.midi_outports.clear()
        self.port = self.client.midi_outports.register('out')
        if not self.port.is_connected_to(carla_ports[0]):
            self.client.connect(self.port, carla_ports[0])
        self.is_active = True

    def wait(self, in_fw=False, out_fw=False):
        """
        waits while setting freewheeling mode to `in_fw`
        it then set freewheeling mode to `out_fw` before exiting
        """
        if hasattr(self, 'end_midiplayer'):
            self.set_freewheel(in_fw)
            self.end_midiplayer.wait()
            self.set_freewheel(out_fw)
        self.deactivate()

    def clear(self):
        """
        clears the `_messages` list
        """
        del self._messages
        self._messages = []

    def synthesize_messages(self,
                            messages: List[mido.Message],
                            sync=False,
                            condition=lambda: True,
                            **kwargs):
        """
        Synthesize a list of messages

        1. Connect the port of this jack client to Carla if not yet done
        2. Send the list of messages to the Carla instance

        If `sync` is True, this function waits until all messages have been
        processed, otherwise, it suddenly returns. You can wait by calling the
        `wait` method of this object.

        This function is compatible with freewheeling mode.  Freewheel prevents
        jack from waiting between return calls. This allows for the maximum
        allowed speed, but not output/input operation is done with system audio
        (i.e. you cannot listen/recording to anything while in freewheeling
        mode).

        `condition` is a function checked in the playing callback. If
        `condition()` is False, no message is sent. The callback start playing
        at the cycle after the one in which `condition()` becomes True.

        `kwargs` are passed to `wait` if `sync` is True.

        Note: Mido numbers channels 0 to 15 instead of 1 to 16. This makes them
        easier to work with in Python but you may want to add and subtract 1
        when communicating with the user.
        """
        self._messages = messages

        global msg, offset
        it = iter(self._messages)
        msg = next(it)
        offset = 0
        self.end_midiplayer = threading.Event()
        self.ready_at = -1

        @self.client.set_process_callback
        def process(frames):
            if self.is_active:
                global offset, msg
                for port in self.client.midi_outports:
                    if not self.is_ready():
                        print(self.client.name + " ready!")
                        # let other clients know that this is ready
                        self.ready_at = self.client.last_frame_time
                    elif condition():
                        # start only if other clients are ready too
                        port.clear_buffer()
                        while True:

                            if offset >= frames:
                                # wait for the correct block
                                offset -= frames
                                return
                            # Note: This may raise an exception:
                            port.write_midi_event(offset, msg.bytes())

                            try:
                                msg = next(it)
                            except StopIteration:
                                self.end_midiplayer.set()
                            offset += round(msg.time * self.client.samplerate)

        self.activate()
        if sync:
            self.wait(**kwargs)

    def synthesize_midi_note(self,
                             pitch: int,
                             velocity: int,
                             duration: float,
                             sustain: int = 0,
                             soft: int = 0,
                             sostenuto: int = 0,
                             channel: int = 0,
                             program: int = 0,
                             **kwargs) -> multiprocessing.Process:
        """
        set up a list of messages representing one note and then calls
        `self.synthesize_messages`. All keywords from that method can be used
        here.
        """
        messages = [
            # program change
            mido.Message('program_change',
                         program=program,
                         channel=channel,
                         time=0),
            # sustain
            mido.Message('control_change',
                         control=64,
                         value=sustain,
                         channel=channel,
                         time=0),
            # sostenuto
            mido.Message('control_change',
                         control=66,
                         value=sostenuto,
                         channel=channel,
                         time=0),
            # soft
            mido.Message('control_change',
                         control=67,
                         value=soft,
                         channel=channel,
                         time=0),
            # note on
            mido.Message('note_on',
                         velocity=velocity,
                         note=pitch,
                         channel=channel,
                         time=0),
            # note off
            mido.Message('note_off',
                         note=pitch,
                         channel=channel,
                         time=duration)
        ]

        return self.synthesize_messages(messages, **kwargs)

    def synthesize_midi_file(self, midifile: Any,
                             **kwargs) -> multiprocessing.Process:
        """
        Send midi messages contained in `filename` using
        `self.synthesize_messages`. All keywords from that method can be used
        here.

        `midifile` can be a `mido.MidiFile` object or a string

        After the playback, ports are resetted
        """

        if type(midifile) is str:
            midifile = mido.MidiFile(midifile)

        messages = [m for m in midifile]

        return self.synthesize_messages(messages, **kwargs)


class ExternalMIDIPlayer(ExternalProcess):

    MIDI_PORT = 'Carla:events-in'

    def __init__(self):
        super().__init__()
        self._tempname = 'tmp.mid'

        # check dependencies
        if shutil.which('jack-smf-player') is None:
            raise RuntimeError(
                "Dependencies not satisfied, install jack-smf-player in your\
PATH")

    def _start_player(self, filename):
        self.process = Popen(
            ['jack-smf-player', '-q', '-a', self.MIDI_PORT, filename])
        return self.process

    def synthesize_midi_note(self,
                             pitch: int,
                             velocity: int,
                             duration: float,
                             sustain: int = 0,
                             soft: int = 0,
                             sostenuto: int = 0,
                             channel: int = 0,
                             program: int = 0):
        """
        Send midi messages representing the note to `MIDI_PORT` and waits for
        it

        The `Popen` object is in `self.process`.
        """
        self.process.wait()
        mid = mido.MidiFile()
        track = mido.MidiTrack()
        # program
        track.append(
            mido.Message('program_change', program=program, channel=channel))
        # sustain
        track.append(
            mido.Message('control_change',
                         control=64,
                         time=0,
                         value=sustain,
                         channel=channel))
        # sostenuto
        track.append(
            mido.Message('control_change',
                         control=66,
                         time=0,
                         value=sostenuto,
                         channel=channel))
        # soft
        track.append(
            mido.Message('control_change',
                         control=67,
                         time=0,
                         value=soft,
                         channel=channel))
        # note on
        track.append(
            mido.Message('note_on',
                         velocity=velocity,
                         time=0,
                         note=pitch,
                         channel=channel))
        # note off
        track.append(
            mido.Message('note_off',
                         note=pitch,
                         time=duration,
                         channel=channel))
        mid.tracks.append(track)
        mid.save(self._tempname)
        self._duration = duration
        self._start_player(self._tempname)
        self.wait()

    def synthesize_midi_file(self,
                             midifile: Union[mido.MidiFile, str],
                             sync=True,
                             progress=True) -> multiprocessing.Process:
        """
        Send midi messages contained in `filename` to `self.port`

        `midifile` can be a `mido.MifiFile` object or a string

        `self.port` is initialized during in the constructor of this object but
        if no Carla instance is found it is automatically updated when
        entering this method. You can force an update with `self._update_port`
        method.

        If `sync` is set to False, a new sub-process is dispached and this
        function returns suddenly.  The `multiprocessing.Process` object is
        returned and you can wait for that process to terminate (e.g. by
        calling its `join` method). You can also wait by calling the `wait`
        method of this object.

        If `sync` is True, a new process is dispatched, but it is waited in
        before returning.

        If `progress` is True, a bar is printed showing the progress of the
        synthesis. If `sync` is False, `progress` is not used.
        """
        self.process.wait()
        if type(midifile) is mido.MidiFile:
            self._duration = midifile.length  # type: ignore
            midifile.save(self._tempname)  # type: ignore
            midifile = self._tempname
        else:
            self._duration = get_smf_duration(midifile)
        self._start_player(midifile)

        if sync:
            _wait_dur(self._duration,
                      progress,
                      condition=lambda: self.process.is_running())
            self.kill()
        return self.process

    def close(self):
        """
        Remove temporary files created by this instance
        """
        if os.path.exists(self._tempname):
            os.remove(self._tempname)


def _wait_dur(dur, progress, condition=lambda: True, q=0.005):
    i = 0.0
    while i < dur and condition():
        if progress:
            progressbar(i, 1, dur, status='Synthesis')
        time.sleep(q)
        i += q
    if progress:
        progressbar(dur, 1, dur, status='Completed!')



import time

import numpy as np
import soundfile as sf

from .utils import JackClient


class AudioRecorder(JackClient):
    AUDIO_PORT = 'Carla'

    def __init__(self):
        """
        Records output from a Carla instance.

        For now, only one Carla instance should be active.

        If the Carla instance is not found, this method raises a
        `RuntimeWarning`. To avoid it, use ``Carla.exists`` method. Note that
        ``Carla.start`` already does that!
        """
        super().__init__("AudioRecorder")

    def activate(self):
        """
        Activate the recording client and set the connections.
        Set self.channels and create one input port per each Carla output port.

        If the Carla instance is not found, this method rase a
        `RuntimeWarning`. To avoid it, use ``Carla.exists`` method. Note that
        ``Carla.start`` already does that!
        """
        # get default values and params
        carla_ports = self.client.get_ports(self.AUDIO_PORT,
                                            is_audio=True,
                                            is_output=True)
        if len(carla_ports) == 0:
            raise RuntimeWarning(
                "Cannot find the Carla instance, Retry later!")
        self.channels = len(carla_ports)
        self.ports = []
        self.client.activate()
        self.client.inports.clear()
        for i, p in enumerate(carla_ports):
            inp = self.client.inports.register(f"in{i}")
            if not inp.is_connected_to(p):
                self.client.connect(p, inp)
            self.ports.append(inp)
        self.is_active = True

    def clear(self):
        """
        Clears the `recorded` array
        """
        del self.recorded
        self.recorded = []

    def start(self,
              duration=None,
              sync=False,
              condition=lambda: True,
              **kwargs):
        """
        Record audio for ``duration`` seconds. Note that this function blocks
        if `sync` is True, otherwise, this returns suddenly and you
        should wait/stop by calling the `wait` method of this object which
        constructs the recorded array in `self.recorded`

        `condition` is a function checked in the recording callback. If
        `condition()` is False, blocks are discarded. The callback start
        recording at the cycle after the one in which `condition()` becomes
        True.

        This function is compatible with Jack freewheeling mode to record
        offline sessions.

        `kwargs` are passed to `wait` if `sync` is True.
        """
        global callback
        self.recorded = []
        self.ready_at = -1

        @self.client.set_process_callback
        def callback(frames):
            channels = [i.get_array() for i in self.client.inports]
            if len(channels) == self.channels:
                if not self.is_ready():
                    # let other clients know that this is ready
                    print(self.client.name + " ready!")
                    self.ready_at = self.client.last_frame_time
                elif condition():
                    # start only if other clients are ready too
                    self.recorded.append(np.stack(channels))

        if duration is not None:
            self._needed_samples = int(duration * self.client.samplerate)
        else:
            # values <= 0 cause the wait to hang until `timeout`
            self._needed_samples = -1

        self.activate()
        if sync:
            self.wait(**kwargs)

    def wait(self, timeout=None, in_fw=False, out_fw=False):
        """
        Wait until recording is finished. If `timeout` is a number, it should
        be the maximum number of seconds until which the recording stops. A
        boolean is returned representing if timeout is reached.
        (returns `False` if timeout is not set)

        The recording stops when `timeout` or the duration passed when calling
        `start` is reached. In these cases, the recording client is deactivated
        and the callback stopped.

        waits while setting freewheeling mode to `in_fw`
        it then set freewheeling mode to `out_fw` before exiting
        """
        assert timeout is not None or self._needed_samples > 0, "Please, provide one between`timeout` and `duration`"
        reached_timeout = False
        if self.is_active:
            self.set_freewheel(in_fw)
            # wait the needed number of blocks
            ttt = time.time()
            CONTINUE = True
            while CONTINUE:
                if timeout is not None:
                    if time.time() - ttt > timeout:
                        reached_timeout = True
                        break
                time.sleep(0.001)
                if self._needed_samples > 0:
                    recorded_frames = sum(i.shape[1] for i in self.recorded)
                    CONTINUE = recorded_frames < self._needed_samples

            self.set_freewheel(out_fw)
            self.deactivate()
            try:
                self.recorded = np.concatenate(self.recorded, axis=1)
            except Exception:
                print(
                    "Cannot concatenate the recorded blocks in one array, have you changed the number of channels while recording? In the `AudioRecorder.recorded` you will find the list of recorded blocks"
                )
        return reached_timeout

    def save_recorded(self, filename):
        """
        Save the recorded array to file. Extensions supported by
        ``libsndfile``!

        `start_frame` is the frame from which recorded is saved (use it to
        discard initial delays due to Jack setup).
        """

        if not hasattr(self, 'recorded'):
            raise RuntimeError("No recorded array!")

        recorded = self.recorded.T
        try:
            sf.write(str(filename), recorded, int(self.client.samplerate))
        except Exception:
            print(
                "Cannot write with default sounddevice parameters. Try writing by yourself using `recorded` field!"
            )



"""
Simple test.

Requires aplay
Use with an argument consisting of the path to a MIDI file
"""
import subprocess
import sys

import mido
import numpy as np

from . import AudioRecorder, Carla, JackServer, MIDIPlayer, get_smf_duration

FINAL_DECAY = 4
server = JackServer(['-R', '-d', 'alsa', '-p', '1024'])
carla = Carla("../../carla_proj/pianoteq0.carxp", server, min_wait=4)
carla.start()

filename = sys.argv[1]
player = MIDIPlayer()
recorder = AudioRecorder()

# testing one note
# print("Playing and recording one note in real-time mode..")
# duration = 1
# pitch = 64
# recorder.start(duration + FINAL_DECAY, sync=False)
# player.synthesize_midi_note(pitch, 64, duration, 0, sync=False)
# recorder.wait()
# audio = recorder.recorded
# if not np.any(audio):
#     print("\nError, no sample != 0")
#     carla.kill()
#     sys.exit()

# testing full midi file
player = MIDIPlayer()
print("Playing and recording full file in freewheeling mode..")
duration = get_smf_duration(filename)
recorder.start(duration + FINAL_DECAY, sync=False)
server.toggle_freewheel()
player.synthesize_midi_file(filename, sync=True, progress=True)
recorder.wait()
server.toggle_freewheel()
recorder.save_recorded("session.wav")
carla.kill()

# playing wav file
print("Playing recorded file")
proc = subprocess.Popen(['play', 'session.wav'])
proc.wait()



import subprocess
import sys
import time
import os

import jack
import psutil
import mido


def get_smf_duration(filename):
    """
    Return note dration of a file from a path
    """
    return mido.MidiFile(filename).length


def progressbar(count, block_size, total, status='Download'):
    bar_len = 60
    count *= block_size
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()


def Popen(command, **args):
    """
    Custom call to ``psutil.Popen``
    """
    return psutil.Popen(command,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                        **args)


def kill_psutil_process(process):
    """
    Kills a tree of `psutil.Process`
    """

    children = process.children(recursive=True)
    children.append(process)
    for p in children:
        try:
            p.kill()
        except psutil.NoSuchProcess:
            pass


def find_procs_by_name(name):
    """Return a list of processes matching 'name'."""
    ls = []
    for p in psutil.process_iter(["name", "exe", "cmdline"]):
        if name == p.info['name'] or \
                p.info['exe'] and os.path.basename(p.info['exe']) == name or \
                p.info['cmdline'] and p.info['cmdline'][0] == name:
            ls.append(p)
    return ls


class JackClient:
    def __init__(self, name):
        self.client = jack.Client(name)
        self.is_active = False
        self.ready_at = -1

    def is_ready(self):
        """
        Check if the client is active,  if it is in condition of processing
        and, if the current cycle is successive to the one in which the client
        became in condition of processing.

        The latter condition is needed to guarantee to all the other clients
        that this function starts returning `True` in the same cycle --
        otherwise, clients that were processed before of this could see this
        client not ready, while clients processed after this see it as ready.
        """
        return self.ready_at < self.client.last_frame_time\
            and self.ready_at > 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if isinstance(exc_type, Exception):
            print(exc_tb)

    def deactivate(self):
        """
        Deactivates the client and unregisters all input ports
        """
        if self.is_active:
            self.client.deactivate()
            self.client.inports.clear()
            self.client.outports.clear()
            self.client.midi_inports.clear()
            self.client.midi_outports.clear()
            self.is_active = False

    def close(self):
        """
        Deactivate, close and clear memory from this client
        """
        self.deactivate()
        self.client.close()
        self.clear()

    def set_freewheel(self, onoff: bool):
        """
        Set freewheel on or off, without raising exceptions
        """
        try:
            self.client.set_freewheel(onoff)
        except jack.JackError:
            print(f"Cannot set freewheel to {onoff}")

    def clear(self):
        pass

    def activate(self):
        raise NotImplementedError("Abstract method")

    def wait(self, in_fw=False, out_fw=False):
        """
        waits while setting freewheeling mode to `in_fw`
        it then set freewheeling mode to `out_fw` before exiting
        """
        raise NotImplementedError("Abstract method")


class FakeProcess:
    """
    A class to create a fake process/popen for initializing
    `ExternalProcess`
    """
    def __init__(obj):
        pass

    def wait(obj):
        pass

    def is_running(obj):
        """
        from psutil.Process
        """
        return False

    def is_alive(obj):
        """
        from multiprocessing.Process
        """
        return False

    def status(obj):
        return 'dead'

    def children(obj, *args, **kwargs):
        return []

    def terminate(obj):
        pass

    def kill(obj):
        pass

    def start(obj):
        pass


class ExternalProcess:
    """
    A class for processes with a pre-defined duration
    """
    def __init__(self, *args):
        self.process = FakeProcess()
        self._duration = 0
        self.args = args

    def kill(self):
        """
        Just calls `self.process.kill()` and reset this object
        """
        for i in range(10):
            self.process.kill()
            if not self.process.is_running() or self.process.status(
            ) == 'zombie':
                self.__init__(*self.args)
                return
            time.sleep(0.5)

        raise Exception("Cannot kill a process:", self)

    def wait(self):
        """
        Wait the `self._duration` and then kill the process.
        """
        if self._duration > 0:
            timeout = None
        else:
            timeout = self._duration

        try:
            self.process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.kill()



from typing import List
import shutil
import time

import jack

from .utils import ExternalProcess, Popen, find_procs_by_name, kill_psutil_process


class JackServer(ExternalProcess):
    def __init__(self, options):
        """
        Starts a jack server with given options and create a dummy client named
        `pycarla` to query and interact with it

        Args
        ----
        `options` : list[str]
            list of options to be passed to Popen
        """
        super().__init__(options)
        self.options = options
        if not shutil.which('jackd'):
            raise Warning(
                "Jack seems not to be installed. Install it and put the \
``jackd`` command in your path.")

    def start(self):
        """
        Starts the server if not already started
        """
        try:
            self.connect()
            self.process = find_procs_by_name('jackd')[0]
        except jack.JackOpenError as e:
            print("Jack server is not running, starting it!")
            if not self.process.is_running():
                self.process = Popen(['jackd'] + self.options)
            else:
                raise e
            time.sleep(1)
            self.connect()

    def connect(self):
        if not hasattr(self, 'client'):
            self.client = jack.Client('pycarla')
        if not hasattr(self, 'client'):
            print("Cannot connect to Jack server!")

    def restart(self):
        """
        Wait for the duration of this `ExternalProcess`, then kill and restart.
        If the duration is not set, it doesn't return
        """
        self.wait()
        self.start()

    def get_ports(self) -> List[str]:
        return [port.name for port in self.client.get_ports()]

    def kill(self):
        self.client.close()
        kill_psutil_process(self.process)
        super().kill()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycarla']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.2,<6.0.0',
 'jack-client>=0.5.2,<0.6.0',
 'mido>=1.2.9,<2.0.0',
 'numpy>=1.19.0,<2.0.0',
 'psutil>=5.8.0,<6.0.0',
 'pysoundfile>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'pycarla',
    'version': '0.3.2',
    'description': 'Use VST/LV2/etc. plugins with realtime abilities in Python',
    'long_description': 'pyCarla\n==========\n\nA python module for synthesizing MIDI events and files\nfrom python code with using any kind of audio plugin!\n\nSee `docs <https://pycarla.readthedocs.org>`_ for more installation and more info.\n\nTLDR\n----\n\nPython has no strong real-time capabilities since it cannot run with parallel threads.\nThis method delegates most of the realtime stuffs to external C/C++ programs, improving\nthe performances and the accuracy against pure-Python based approaches.\n\nThis method is really portable and supports almost any type of plugins and\nvirtual instruments thanks to the excellent Carla:\n\n#. Linux VST2/VST3\n#. Windows VST2/VST3\n#. LV2\n#. LADSPA\n#. DSSI\n#. AU\n#. SF2/SF3\n#. SFZ\n#. Any other format supported by external plugins\n\nTODO\n----\n\nFor sure\n~~~~~~~~\n\n#. Add single function to synthesize midi file\n#. Add single function to batch-synthesize midi files\n#. Make synchronization handling xruns (edit `is_ready`)\n#. Update Carla\n#. refactoring with a generic `start` method in `JackClient` and \n   specific methods `get_callback` and `get_duration`\n\nMaybe\n~~~~~\n\n#. Use Carla python code to control Carla host\n#. Support LADISH sessions for automatically starting LinuxSampler if needed\n#. Installation scripts for windows and mac\n\n\nCredits\n=======\n\n#. `Federico Simonetta <https://federicosimonetta.eu.org>`_\n    ``federico.simonetta`` ``at`` ``unimi.it``\n',
    'author': 'Federico Simonetta',
    'author_email': 'federico.simonetta@unimi.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pycarla.readthedocs.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import argparse
import fnmatch
import os
import platform
import random
import shutil
import signal
import subprocess
import sys
import tarfile
import time
import urllib.request

import mido

import psutil
from .jackserver import JackServer
from .utils import ExternalProcess, FakeProcess, progressbar, kill_psutil_process

version = "2.1"


def download():

    if sys.platform == 'linux':
        # download carla
        print("Downloading...")
        arch = platform.architecture()[0][:2]
        target_url =\
            'https://github.com/falkTX/Carla/releases/download/v' + \
            version + '/Carla_' + version + '-linux' +\
            arch + '.tar.xz'
        try:
            fname, header = urllib.request.urlretrieve(target_url,
                                                       reporthook=progressbar)
        except Exception as e:
            print("Error while downloading Carla!", file=sys.stderr)
            print(e, file=sys.stderr)
        progressbar(10, 1, 10, status='Completed!')
        print("\n")

        print("Extracting archive...")
        rand = str(random.randint(10**4, 10**5))
        tmp_path = "/tmp/carla" + "-" + rand
        with tarfile.open(fname, "r:xz") as file:
            file.extractall(tmp_path)

        shutil.move(tmp_path + "/Carla_" + version + "-linux" + arch,
                    CARLA_PATH)
    else:
        print(
            "Your OS is against the freedom of software developers because its\
sources are closed. I don't support closed software. You can still download\
Carla by yourself and put the `Carla` command in your path.")


def run_carla():
    server = JackServer(['-R', '-d', 'alsa'])
    carla = Carla("", server, min_wait=0, nogui=False)
    carla.start()
    carla.process.wait()
    try:
        carla.kill()
    except:
        print("Processes already closed!")


mido.set_backend('mido.backends.rtmidi/UNIX_JACK')

THISDIR = os.path.dirname(os.path.realpath(__file__))

if sys.platform == 'linux':
    # use our carla version
    CARLA_PATH = os.path.join(THISDIR, "carla") + "/"
    # from pathlib import Path
    # CARLA_PATH = Path(THISDIR) / "source" / "frontend" / "carla"
else:
    # use the system version
    CARLA_PATH = ""


class Carla(ExternalProcess):
    def __init__(self,
                 proj_path: str,
                 server: JackServer,
                 min_wait: float = 0,
                 nogui: bool = True):
        """
        Creates a Carla object, ready to be started.

        `min_wait` is the minimum amount of seconds waited when starting Carla;
        it is useful if your preset takes a bit to be loaded.

        `nogui` is False if you want to use the gui
        """
        super().__init__()

        self.proj_path = proj_path
        self.server = server
        self.min_wait = min_wait
        if nogui:
            self.nogui = "-n"
        else:
            self.nogui = ""

        if sys.platform == 'linux':
            if not os.path.exists(CARLA_PATH):
                raise Warning("Carla seems not to be installed. Run \
``python -m pycarla.carla -d`` to install it!")
        else:
            if not shutil.which('carla'):
                raise Warning(
                    "Carla seems not to be installed. Download it and put the \
``Carla`` command in your path.")

    def restart_carla(self):
        """
        Only restarts Carla, not the Jack server!
        """
        self.kill_carla()
        self.start()

    def restart(self):
        """
        Restarts both the server and Carla!
        """
        print("Restarting Carla")
        self.kill_carla()
        self.server.kill()
        self.start()

    def __make_carla_popen(self, proj_path):
        self.process = psutil.Popen(
            [CARLA_PATH + "Carla", self.nogui, proj_path],
            preexec_fn=os.setsid)

    def start(self):
        """
        Start carla and Jack and wait `self.min_wait` seconds after a Carla
        instance is ready.
        """
        self.server.start()

        if self.proj_path:
            proj_path = os.path.abspath(self.proj_path)
        else:
            proj_path = ""

        # starting Carla
        self.__make_carla_popen(proj_path)

        # waiting
        start = time.time()
        READY = self.exists()
        while True:
            if not READY and self.exists():
                start = time.time()
                READY = True
            if time.time() - start >= self.min_wait and READY:
                break
            if time.time() - start >= 20:
                self.kill_carla(sign=signal.SIGKILL)
                self.server.restart()
                self.__make_carla_popen()
                start = time.time()
            time.sleep(0.1)

    def kill_carla(self):
        """
        kill carla, but not the server
        """
        kill_psutil_process(self.process)

    def kill(self):
        """
        kill carla and wait for the server
        """
        for i in range(10):
            self.kill_carla()
            if not self.exists():
                break
            time.sleep(0.5)
        self.server.kill()

    def exists(self, ports=["Carla:events*", "Carla:audio*"]):
        """
        simply checks if the Carla process is running and ports are available

        `ports` is a list of string name representing Jack ports; you can use
            '*', '?' etc.

        Returns
        ---
        bool: True if all ports in `ports` exist and the Carla process is
            running, false otherwise
        """
        real_ports = self.server.get_ports()
        for port in ports:
            if not fnmatch.filter(real_ports, port):
                return False

        if not self.process.is_running():
            return False

        if self.process.status() == 'zombie':
            print("Warning! Carla is Zombie!")
            self.process = FakeProcess()
            return False

        return True

    def wait_exists(self):
        """
        Waits until a Carla instance is ready in Jack
        """
        while not self.exists():
            time.sleep(0.5)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="Manage the Carla instance verion")
    argparser.add_argument(
        "-d",
        "--download",
        action="store_true",
        help=
        "Download the correct Carla version in the installation directory of\
    this python package.")

    argparser.add_argument(
        "-r",
        "--run",
        action="store_true",
        help="Run the Carla instance previously downloaded.")

    args = argparser.parse_args()
    if args.download and args.run:
        print("Please, provide one option at a time!")
    else:
        if args.download:
            download()
        elif args.run:
            run_carla()
        else:
            argparser.print_help()
